/**
* @file  df_api.cpp
* @brief Zentrales Modul der DFComDLL.
*/
//////////////////////////////////////////////////////////////////////

#include "api/df_api.h"
#include "datafox_library_iv.h"
#include "dfcom_version.h"

#if defined(_MSC_VER) || defined(__MINGW32__)
#define snprintf	    _snprintf
#endif

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/foreach.hpp>
#include <string>
#include <set>
#include <exception>
#include <sstream>
#include <regex>

/*
"device.serial";
"device.type";
// Ermitteln der Hardware
"modul.board");
"modul.rs485";
"modul.tcpip";
"modul.modemMC";
"modul.rs485ZK";
"modul.transponder";
"modul.display";
"modul.keypad";
"modul.mmc";
"modul.gps";
"modul.usb";
"modul.radio";
"modul.dcf77";
"modul.inAnalog";
"modul.inDigital";
"modul.outDigital";
"modul.fingerprint";
"modul.externalReader";
"option.ioModule";
"option.tbAdaption");
"option.comServer";
"option.tpWrite";
"option.http";
"option.language";
"option.biokeyUpdate";
"option.setup";
"modul.touch";
"modul.sensor";
"modul.battery";
"modul.charge";	
*/

namespace {

class ini_struct
{
public:
    ini_struct()
    {
        // Alle Keys in Map halten
        setting.insert("serial");
        //    setting.insert("version");
        setting.insert("type");
        // Ermitteln der Hardware
        setting.insert("board");
        //    setting.insert("rs485");
        //    setting.insert("tcpip");
        //    setting.insert("modemMC");
        setting.insert("rs485ZK");
        setting.insert("transponder");
        //    setting.insert("display");
        //    setting.insert("keypad");
        //    setting.insert("mmc");
        //    setting.insert("gps");
        //    setting.insert("usb");
        //    setting.insert("radio");
        //    setting.insert("dcf77");
        //    setting.insert("inAnalog");
        //    setting.insert("inDigital");
        //    setting.insert("outDigital");
        //    setting.insert("fingerprint");
        setting.insert("externalReader");
        //    setting.insert("ioModule");
        //    setting.insert("tbAdaption");
        //    setting.insert("comServer");
        //    setting.insert("tpWrite");
        //    setting.insert("http");
        //    setting.insert("language");
        //    setting.insert("fingerprintUpdate");
        //    setting.insert("setup");
        //    setting.insert("touch");
        //    setting.insert("sensor");
        //    setting.insert("battery");
        //    setting.insert("charge");	
    }

public:
    std::set<std::string> setting;
};

}

// ########################################################
// #####################   EXPORTE    #####################
// ########################################################

#if defined(_MSC_VER) || defined(__MINGW32__)
// ########################################################
/**
* @fn    CALLBACK DllGetVersion(DLLVERSIONINFO *pdvi)
* @brief Ermitteln der DLL - Version. Kompatible zu Windows - Standard.
*
* @param pdvi       { Zeiger auf Struktur zur Informationsuebergabe. }
*/
DF_API_EXTERN HRESULT DF_API_CALL DllGetVersion( DLLVERSIONINFO *pdvi )
{
    TCHAR szModuleName[MAX_PATH];
    DWORD dwHandle;
    DWORD dwLen;
    LPVOID lpBuffer = nullptr;
    UINT uiLen;

    unsigned char *pucData = nullptr;

    // zulaessige Bereiche pruefen
    if ( pdvi == nullptr )
    {
        return E_ABORT;
    }

    if ( pdvi->cbSize != sizeof(DLLVERSIONINFO) )
    {
        return E_ABORT;
    }

    // Modulname der DLL
    if ( !GetModuleFileName( theApp._hModule, szModuleName, MAX_PATH ) )
    {
        return E_ABORT;
    }

    // Laenge der Version - Resource
    dwLen = GetFileVersionInfoSize( szModuleName, &dwHandle );
    if ( !dwLen )
    {
        return E_ABORT;
    }

    pucData = new unsigned char[dwLen + 100]; // unicode
    // DLL - Versioninfo ermitteln
    if ( !GetFileVersionInfo( szModuleName, dwHandle, dwLen, (LPVOID)pucData ) )
    {
        delete pucData;
        return E_ABORT;
    }

    // Pointer auf allgemeine Infostruktur ermitteln
    if ( !VerQueryValue( (LPVOID)pucData, TEXT("\\\0"), &lpBuffer, &uiLen ) )
    {
        delete pucData;
        return E_ABORT;
    }

    // Daten zurueckgeben
    VS_FIXEDFILEINFO *pvs_ffi = (VS_FIXEDFILEINFO*)lpBuffer;
    pdvi->dwMajorVersion = pvs_ffi->dwFileVersionMS >> 16;
    pdvi->dwMinorVersion = pvs_ffi->dwFileVersionMS & 0x0000FFFF;
    pdvi->dwBuildNumber  = pvs_ffi->dwFileVersionLS >> 16;
    pdvi->dwPlatformID   = DLLVER_PLATFORM_WINDOWS;

    delete pucData;
    return NOERROR;
}
#endif

DF_API_EXTERN_DECORATE int DF_API_CALL DFCVersionLibrary( char *value, int valueLength )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if (channelBridge == nullptr) 
    {
        return 0;
    }

    channelBridge->logBeginPure( __func__, "" );

    // zulaessige Bereiche pruefen
    if ( value == nullptr || valueLength < 12 )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%p\t%i\t(0, %i)", value, valueLength, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    strncpy( value, g_dfcomVersion, valueLength - 1 );

    channelBridge->logEnd( "\t%s\t%i\t(1, %i)", value, valueLength, channelBridge->getLastErrorNumber( 0 ) );

    return 1;
}

DF_API_EXTERN_DECORATE void DF_API_CALL DFCGetErrorText( int connectionNumber, int errorNumber, int languageId, char *value, int valueLength )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return;
    }

    channelBridge->logBegin( __func__, "\t%i\t%i", errorNumber, languageId );

    // zulaessige Bereiche pruefen
    if ( value == nullptr )
    {
        channelBridge->logEnd( "\t\t%i\t(-, %i)", valueLength, IV_ERROR_FUNCTION_PARAMETER );
        return;
    }

    switch ( errorNumber )
    {
    case IV_ERROR_CRC_PACKET:
        snprintf( value, valueLength, "CRC Fehler." );
        break;

    case IV_ERROR_CHANNEL_OPEN:
        snprintf( value, valueLength, "Oeffnen der Schnittstelle fehlgeschlagen." );
        break;

    case IV_ERROR_CHANNEL_CLOSED:
        snprintf( value, valueLength, "Schnittstelle ist geschlossen." );
        break;

    case IV_ERROR_CHANNEL_COMMUNICATION:
        snprintf( value, valueLength, "Allgemeiner Fehler waehrend Kommunikation." );
        break;

    case IV_ERROR_RECEIVE:
        snprintf( value, valueLength, "Geraet antwortet nicht." );
        break;

    case IV_ERROR_TRANSMIT:
        snprintf( value, valueLength, "Fehler waehrend Senden eingetreten." );
        break;

        /*
        case ERROR_AT_RECIEVING	:
        snprintf( szText, iLength, "Fehler waehrend Empfang eingetreten." );
        break;
        */

    case IV_ERROR_ADDRESS:
        snprintf( value, valueLength, "Antwort von falschem Geraet oder Antwort nicht sinnvoll." );
        break;

    case IV_ERROR_DEVICE_ID_REMOTE:
        snprintf( value, valueLength, "Keine gueltige Busnummer. (1 - 31, 254)" );
        break;

    case IV_ERROR_CHANNEL:
        snprintf( value, valueLength, "Kanal nicht vorhanden oder nicht erstellbar." );
        break;

    case IV_ERROR_FILE_OPEN:
        snprintf( value, valueLength, "Oeffnen einer Datei gescheitert." );
        break;

    case IV_ERROR_FILE_FORMAT:
        snprintf( value, valueLength, "z.B. falsches Format einer Setupdatei." );
        break;

    case IV_ERROR_LIST_PROCESSING:
        snprintf( value, valueLength, "Fehler bei Listendatenverarbeitung." );
        break;

    case IV_ERROR_VERSION_FIRMWARE:
        snprintf( value, valueLength, "Firmwareversion nicht kompatible zu Routine." );
        break;

    case IV_ERROR_DATA_AVAILABLE:
        snprintf( value, valueLength, "Es sind noch Daten im Geraet vorhanden." );
        break;

    case IV_ERROR_USER_CANCEL:
        snprintf( value, valueLength, "Durchfuehrung wurde von Anwender abgebrochen." );
        break;

    case IV_ERROR_DESCRIPTION_RECORD_UNLOADED:
        snprintf( value, valueLength, "Keine Datensatzbeschreibung geladen." );
        break;

    case IV_ERROR_DESCRIPTION_LIST_UNLOADED:
        snprintf( value, valueLength, "Keine Listenbeschreibung geladen." );
        break;

        /*
        case ERROR_FLASH_PROG:
        snprintf( szText, iLength, "Fehler beim Programmieren des Flash, Firmware." );
        break;
        */

    case IV_ERROR_FUNCTION_PARAMETER:
        snprintf( value, valueLength, "Parameter einer Funktion ausserhalb ihres Gueltigkeitsbereiches." );
        break;

    case IV_ERROR_VARIABLE_UNDEFINED:
        snprintf( value, valueLength, "Globale Variable nicht definiert." );
        break;

    case IV_ERROR_VARIABLE_SIZE:
        snprintf( value, valueLength, "Der verwendete Variablenwert ist zu lang." );
        break;

    case IV_ERROR_VARIABLE_TYPE:
        snprintf( value, valueLength, "Der verwendete Variablentype wuerde nicht eingehalten." );
        break;

    case IV_ERROR_VARIABLE_RIGHT:
        snprintf( value, valueLength, "Das Schreib-/Leserecht wuerde nicht eingehalten." );
        break;

    case IV_ERROR_VARIABLE_FORMAT_OR_RANGE:
        snprintf( value, valueLength, "Das verwendete Format oder der Wertebereich wuerde nicht eingehalten." );
        break;

        /*
        case ERROR_NO_DATA:
        snprintf( szText, iLength, "Anfrage obwohl keine Daten." );
        break;
        */

        /*
        case ERROR_MULTI:
        snprintf( szText, iLength, "Dublikat von Datensaetzen abgefangen." );
        break;
        */

    case IV_ERROR_BOOTVERSION:
        snprintf( value, valueLength, "Geraetedatei nicht mit Bootloaderversion kompatible." );
        break;

        /*
        case ERROR_BOOTFILEVERSION:
        snprintf( szText, iLength, "Firmwaredatei nicht mit Bootloaderversion kompatible." );
        break;
        */


    case IV_ERROR_FUNCTION_UNAVAILABLE:
        snprintf( value, valueLength, "Funktion steht im vorliegenden Fall nicht zur Verfuegung." );
        break;

    case IV_ERROR_PROCESSING:
        snprintf( value, valueLength, "Der Befehl wurde erfolgreich uebermittelt, konnte aber nicht durchgefuehrt werden." );
        break;

    case IV_ERROR_DEVICE_UNKNOWN:
        snprintf( value, valueLength, "Der Geraetetyp ist unbekannt. Update der Software erforderlich." );
        break;

    case IV_ERROR_DEVICEFILE_WRONG:
        snprintf( value, valueLength, "Die angegebene Geraetedatei stimmt nicht mit der vorliegenden Hardware ueberein." );
        break;

    case IV_ERROR_DEVICE_WRONG:
        snprintf( value, valueLength, "Geraetetyp stimmt nicht. z.B. Versuchtes Einspielen von Setupdatei fuer PZE-MasterIV in BDE-MasterIV." );
        break;

    case IV_ERROR_FILE_EXTENSION:
        snprintf( value, valueLength, "Die vorliegende Dateiendung wird nicht unterstuetzt." );
        break;

    case IV_ERROR_ENTRANCE2_OFF:
        snprintf( value, valueLength, "Die Zutrittskontrolle ist deaktiviert." );
        break;

    case IV_ERROR_NAME_UNKNOWN:
        snprintf(value, valueLength, "Der Tabellenname oder Spaltenname ist undefiniert." );
        break;

    case IV_ERROR_HANDLE:
        snprintf( value, valueLength, "Ein angegebener Handle ist nicht gueltig." );
        break;

    case IV_ERROR_NULL_RESULT:
        snprintf( value, valueLength, "Der anzuwendende Filter wuerde keine Datensaetze liefert." );
        break;

    case IV_ERROR_BOF:
        snprintf( value, valueLength, "Sprung fuehrte zum Anfang der Tabelle. (Base of File)" );
        break;

    case IV_ERROR_EOF:
        snprintf( value, valueLength, "Sprung fuehrte zum Ende der Tabelle. (End of File)" );
        break;

    case IV_ERROR_VALUE:
        snprintf( value, valueLength, "Der angegebene Feldwert ist Fehlerhaft." );
        break;

    case IV_ERROR_OVERFLOW:
        snprintf( value, valueLength, "Bei einer Datenuebertragung ist ein Speicherueberlauf eingetreten." );
        break;

    case IV_ERROR_NO_DATA_AVAILABLE:
        snprintf( value, valueLength, "Es liegen keine Daten vor." );
        break;

    case IV_ERROR_PROCESSING_DLL:
        snprintf( value, valueLength, "Der Befehl konnte zwar uebermittelt werden, bei den Daten liegt jedoch ein Fehler vor." );
        break;

    case IV_ERROR_MODULE_BUSY:
        snprintf( value, valueLength, "Das angesprochene Modul ist zur Zeit durch einen anderen Process blockiert." );
        break;

    case IV_ERROR_FILE_LOAD_XML:
        snprintf( value, valueLength, "Fehler beim Laden der XML-Datei zu modifyStudioFile." );
        break;

    case IV_ERROR_FILE_UNKNOWN_XML:
        snprintf( value, valueLength, "Unbekannte Elemente der XML-Datei zu modifyStudioFile." );
        break;

    case IV_ERROR_FILE_CONTENT_XML:
        snprintf( value, valueLength, "Fehler im Dateninhalt der XML-Datei zu modifyStudioFile." );
        break;

    case IV_ERROR_FILE_MODIFY:
        snprintf( value, valueLength, "Fehler bei Uebernahme (Anwendung) der zu modifizierenden Daten" );
        break;

    case IV_ERROR_ALLOC_MEM:
        snprintf( value, valueLength, "Speicherallokation konnte nicht durchgefuehrt werden." );
        break;

    case IV_ERROR_VERSION_DEVICEFILE:
        snprintf( value, valueLength, "Diese Version der Geraetedatei kann aufgrund ihrer Aktualitaet nicht verarbeitet werden." );
        break;

    case IV_ERROR_FILE_PROCESSING:
        snprintf( value, valueLength, "Bei der Verarbeitung der Dateidaten trat ein Fehler auf." );
        break;

    case IV_ERROR_DEVICEFILE_MATCH:
        snprintf( value, valueLength, "Die vorliegende Geraetedatei passt nicht zur vorliegenden Hardware. Oder aber: Im Archiv wurde keine passende Geraetedatei gefunden." );
        break;

    case IV_ERROR_DEVICEFILE_LOAD:
        snprintf( value, valueLength, "Fehler beim Laden der Geraetedatei(archive) aufgetreten." );
        break;

    case IV_ERROR_AC_STARTED:
        snprintf( value, valueLength, "Die Aktive Verbindung ist gestartet." );
        break;

    case IV_ERROR_AC_STOPPED:
        snprintf( value, valueLength, "Die Aktive Verbindung ist nicht gestartet." );
        break;

    case IV_ERROR_AC_PORTINUSE:
        snprintf( value, valueLength, "Der angegebene Listensocket ist bereits in Verwendung." );
        break;

    case IV_ERROR_AC_NO_NOTIFY_RECORD:
        snprintf( value, valueLength, "Die Aktive Meldung von Datensaetzen ist deaktiviert." );
        break;

    case IV_ERROR_AC_LISTENER:
        snprintf( value, valueLength, "Allgemeiner Fehler beim Starten der Aktiven Verbindung." );
        break;

    case IV_ERROR_NEED_PASSWORD:
        snprintf( value, valueLength, "Das angegebene Passwort stimmt nicht mit dem benoetigten ueberein. Es wird ein Passwort gefordert." );
        break;

    case IV_ERROR_PASSWORD_LOCK:
        snprintf( value, valueLength, "Geraet ist aufgrund von fehlerhaften Passworteingaben fuer 60 Sekunden gesperrt." );
        break;

    case IV_ERROR_FILE_WRITE:
        snprintf( value, valueLength, "Das Schreiben in die Datei ist fehlgeschlagen." );
        break;

    case IV_ERROR_FILE_READ:
        snprintf( value, valueLength, "Das Lesen aus der Datei ist fehlgeschlagen." );
        break;

    case IV_ERROR_FILE_CRC:
        snprintf( value, valueLength, "Fehlerhafter Dateninhalt. Die Pruefung auf die Datenkonsistenz per CRC schlug fehl." );
        break;

    case IV_ERROR_MEMORY_FULL:
        snprintf( value, valueLength, "Es steht kein weiterer Speicherplatz zur Verfuegung.\nZ. B. kann die maximale aufnehmbare Anzahl Fingertemplates erreicht sein." );
        break;

    case IV_ERROR_FILE_DATA_FORMAT:
        snprintf( value, valueLength, "Die in der Datei vorliegenden Daten stimmen nicht mit den geforderten ueberein.\nZ. B. in der Datei liegen Fingertemplates vom Typ Idencom-Compact vor, es werden jedoch DIN Erfordert." );
        break;

    case IV_ERROR_NOT_ENOUGH_FLASH:
        snprintf( value, valueLength, "Es steht nicht genuegend Flash-Speicher zur Verfuegung, um die Daten aufzunehmen." );
        break;

    case IV_ERROR_NO_MODULE:
        snprintf( value, valueLength, "Das von Ihnen angesprochene Modul ist eine Optionale Hardware-Erweiterung und ist im vorliegenden Geraet nicht verfuegbar." );
        break;

    case IV_ERROR_VERSION_SETUPFILE:
        snprintf( value, valueLength, "In der Setupdatei werden Erweiterungen verwendet, die einen aktuelleren Firmwarestand voraussetzen." );
        break;

    case IV_ERROR_WRONG_FIRMWARE:
        snprintf( value, valueLength, "Die abzuarbeitende Aufgabe wird durch die vorliegenden Firmwareversion nicht unterstuetzt." );
        break;

    case IV_ERROR_WRONG_CONFIG:
        snprintf( value, valueLength, "Die abzuarbeitende Aufgabe wird durch die vorliegenden Geraetekonfiguration nicht unterstuetzt." );
        break;

    case IV_ERROR_TIMEBOY_I_II_III:
        snprintf( value, valueLength, "Bei dem angesprochenen Timeboy handelt es sich um kein Model der vierer Reihe." );
        break;

    case IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION:
        snprintf( value, valueLength, "Die Setupdatei setzt eine nicht erreichte Mindestversion der DFComDLL voraus." );
        break;

    case IV_ERROR_SETUPFILE_SHOULD_LIBRARY_VERSION:
        snprintf( value, valueLength, "Die Setupdatei gibt eine nicht erreichte Soll-Version der DFComDLL vor." );
        break;

    case IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE1_VERSION:
        snprintf( value, valueLength, "Die Setupdatei wird von der vorliegenden Firmwareversion 04.01.xx nicht unterstuetzt." );
        break;

    case IV_ERROR_SETUPFILE_MIN_FIRMWARE1_VERSION:
        snprintf( value, valueLength, "Die Setupdatei setzt eine nicht erreichte Mindestversion der Firmware 04.01.xx voraus." );
        break;

    case IV_ERROR_SETUPFILE_SHOULD_FIRMWARE1_VERSION:
        snprintf( value, valueLength, "Die Setupdatei gibt eine nicht erreichte, jedoch empfohlene, Firmwareversion 04.01.xx vor." );
        break;

    case IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE2_VERSION:
        snprintf( value, valueLength, "Die Setupdatei wird von der vorliegenden Firmwareversion 04.02.xx nicht unterstuetzt." );
        break;

    case IV_ERROR_SETUPFILE_MIN_FIRMWARE2_VERSION:
        snprintf( value, valueLength, "Die Setupdatei setzt eine nicht erreichte Mindestversion der Firmware 04.02.xx voraus." );
        break;

    case IV_ERROR_SETUPFILE_SHOULD_FIRMWARE2_VERSION:
        snprintf( value, valueLength, "Die Setupdatei gibt eine nicht erreichte, jedoch empfohlene, Firmwareversion 04.02.xx vor." );
        break;

    case IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE3_VERSION:
        snprintf( value, valueLength, "Die Setupdatei wird von der vorliegenden Firmwareversion 04.03.xx nicht unterstuetzt." );
        break;

    case IV_ERROR_SETUPFILE_MIN_FIRMWARE3_VERSION:
        snprintf( value, valueLength, "Die Setupdatei setzt eine nicht erreichte Mindestversion der Firmware 04.03.xx voraus." );
        break;

    case IV_ERROR_SETUPFILE_SHOULD_FIRMWARE3_VERSION:
        snprintf( value, valueLength, "Die Setupdatei gibt eine nicht erreichte, jedoch empfohlene, Firmwareversion 04.03.xx vor." );
        break;

    case IV_ERROR_FINGERPRINT:
        snprintf( value, valueLength, "Das Fingerprintmodul bringt einen nicht naeher spezifizierten Fehler." );
        break;

    case IV_ERROR_FINGERPRINT_MEMORY_FULL:
        snprintf( value, valueLength, "Die Datenbank des Fingerprintmoduls ist voll." );

        break;
    case IV_ERROR_FINGERPRINT_TEMPLATES_NONE:
        snprintf( value, valueLength, "Das Fingerprintmodul enthaelt keine Fingertemplates." );
        break;

    case IV_ERROR_FINGERPRINT_TEMPLATE_EXISTS:
        snprintf( value, valueLength, "Das Fingertemplate existiert bereits." );
        break;

    case IV_ERROR_FINGERPRINT_TEMPLATE_NOT_EXISTS:
        snprintf( value, valueLength, "Das angeforderte Fingertemplate existiert nicht." );
        break;

        // Fehlernummern für Dateizugriff auf dem Gerät.
    case IV_ERROR_FILE_NOT_FOUND:
        snprintf( value, valueLength, "Die angeforderte Datei existiert nicht." );
        break;

    case IV_ERROR_FILE_WILL_BE_CREATED:
        snprintf( value, valueLength, "Die angeforderte Datei wird aktuell noch erstellt." );
        break;

    case IV_ERROR_MODULE_ABSENCE:
        snprintf( value, valueLength, "Das anzusprechende Hardwaremodul (Zutrittsbus) ist nicht vorhanden." );
        break;

    case IV_ERROR_NO_UUZ_MODULE:
        snprintf( value, valueLength, "Kein U&Z Funkmodul im Zutrittsbus vorhanden." );
        break;

    case IV_ERROR_FW_HEX_FAILED_INFO:
    {
        snprintf( value, valueLength, "Die vorliegende HEX-Datei kann aufgrund fehlender Informationen nicht eingespielt werden. Verwenden Sie bitte die dazugehörige DFZ-Datei." );
        break;
    }

     default:
        snprintf( value, valueLength, "Unbekannter Fehler." );
        break;
    }

    channelBridge->logEnd( "\t%s\t%i\t(-, %i)", value, valueLength, IV_ERROR_SUCCESS );
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCCheckDevice( int connectionNumber, int deviceAddress, int *errorNumber, int devicePollRetry )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || devicePollRetry < 1 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t%i\t(0, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, devicePollRetry, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t%i\t(0, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, devicePollRetry, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->checkDevice( l_device_address, &l_error_number, devicePollRetry );
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
    }

    channelBridge->logEnd( "\t%i\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, devicePollRetry, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetDevicePollRetry( int connectionNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, "" );

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logEnd( "\t(%i, %i)", channelBridge->m_device_poll_retry, channelBridge->getLastErrorNumber( 0 ) );

    return channelBridge->m_device_poll_retry;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComSendMessage( int connectionNumber, int deviceAddress, unsigned char visibleTime, unsigned char doAction, unsigned char audibleSignal, const char *text, int textLength )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t%i\t%i\t%s\t%i", visibleTime, doAction, audibleSignal, (text == nullptr) ? "" : text, textLength );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || (doAction != 0 && doAction != 1) 
        || textLength < 0 
        || (doAction == 0 && textLength > 242)
        || (doAction == 1 && textLength > (IV_PACKET_LENGTH_COMMAND_PARAMS - 3))
        || text == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int iError = IV_ERROR_SUCCESS;
    std::string lText( text, textLength );
    int iRes = channelBridge->sendMessage( l_device_address, visibleTime, doAction, audibleSignal, lText, &iError );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, iError );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );
    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComSendInfotext( int connectionNumber, int deviceAddress, const char *text, int textLength )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%s\t%i", (text == nullptr) ? "" : text, textLength );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || textLength < 0 || textLength > 242 || text == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int iError = IV_ERROR_SUCCESS;
    std::string lText( text, textLength );
    int iRes = channelBridge->sendInfotext( l_device_address, lText, &iError );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, iError );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );
    return iRes;
}


DF_API_EXTERN_DECORATE int DF_API_CALL DFCComSetTime( int connectionNumber, int deviceAddress, const unsigned char *dateTimeBuffer )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->setDateTime( l_device_address, dateTimeBuffer, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    // Uhrzeit mitloggen
    if ( dateTimeBuffer != nullptr )
    {
        channelBridge->logEnd( "\t%02i%02i-%02i-%02i %02i:%02i:%02i\t(%i, %i)",
            dateTimeBuffer[0], dateTimeBuffer[1], dateTimeBuffer[2], dateTimeBuffer[3],
            dateTimeBuffer[4], dateTimeBuffer[5], dateTimeBuffer[6], 
            iRes, channelBridge->getLastErrorNumber( l_device_address ) );
    }
    else
    {
        channelBridge->logEnd( "\t\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );
    }

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComGetTime( int connectionNumber, int deviceAddress, unsigned char *dateTimeBuffer )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || dateTimeBuffer == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t\t(0, %i)", channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->getDateTime( l_device_address, dateTimeBuffer, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    if ( iRes )
    {
        channelBridge->logEnd( "\t%02i%02i-%02i-%02i %02i:%02i:%02i\t(%i, %i)",
            dateTimeBuffer[0], dateTimeBuffer[1], dateTimeBuffer[2], dateTimeBuffer[3],
            dateTimeBuffer[4], dateTimeBuffer[5], dateTimeBuffer[6], 
            iRes, channelBridge->getLastErrorNumber( l_device_address ) );
    }
    else
    {
        channelBridge->logEnd( "\t\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );
    }

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCComVersion( int connectionNumber, int deviceAddress, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->getProtocolVersion2( l_device_address, TRUE, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(0x%04X, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );
    // Kommunikationsversion z.B: 0x00000103
    // Die Versionsnummer ist ein 16Bit - Wert
    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetVersionFirmware( int connectionNumber, int deviceAddress, char *value, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || value == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    std::string lVersion;
    int iRes = channelBridge->getVersionFirmware( l_device_address, lVersion );
    if ( iRes != IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        iRes = channelBridge->resultToError( iRes, &l_error_number );
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
    }
    else
    {
        lVersion.copy( value, lVersion.size() );
        value[lVersion.size()] = 0;
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (iRes) ? value : "", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetVersionFirmwareFromFile( int connectionNumber, const char *filePath, char *value, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, "" );

    // zulaessige Bereiche pruefen
    if ( filePath == nullptr || value == nullptr || strlen( filePath ) < 1 )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS, errorNumber );

    int l_error_number = IV_ERROR_SUCCESS;
    std::string version;

    int iRes = channelBridge->getVersionFirmwareFromFile( filePath, version, &l_error_number );
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( 0, l_error_number, errorNumber );
    }

    std::strcpy( value, version.c_str() );
    channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", filePath, (iRes) ? value : "", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetSeriennummer( int connectionNumber, int deviceAddress, int *errorNumber, int *value )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || value == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->getSerialNum( l_device_address, value, &l_error_number );

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, *value, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCStartTransparencyMode( int connectionNumber, int deviceAddress, int id, int flags, unsigned int sizeBufferRead, unsigned int sizeBufferWrite, int timeout, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t%i\t%i\t%i", id, flags, sizeBufferRead, sizeBufferWrite );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->startTransparencyMode( l_device_address, id, flags, sizeBufferRead, sizeBufferWrite, timeout, &l_error_number );

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCStopTransparencyMode( int connectionNumber, int deviceAddress, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->stopTransparencyMode( l_device_address, &l_error_number );

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCIsTransparencyMode( int connectionNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, "" );

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    int iRes = channelBridge->isTransparencyMode();

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCPing( int connectionNumber, int deviceAddress, int attempts, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", attempts );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || attempts < 1 || attempts > 1000 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->ping( l_device_address, attempts, &l_error_number );

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE void DF_API_CALL DFCSetCallBack( int connectionNumber, int (*functionPointer)() )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return;
    }

    channelBridge->logBegin( __func__, "" );

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->setCallBackFunction( functionPointer );

    channelBridge->logEnd( "" );
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetGlobVar( int connectionNumber, int deviceAddress, const char *name, int nameType, const char *value, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || name == nullptr || value == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t\t%i\t\t%i\t(%i, %i)", nameType, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Bereich pruefen.
    switch ( nameType )
    {
    case 0: // Ansprechen ueber Nummer
        if ( *name < 1 || *name > 8 )
        {
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }

        channelBridge->logContinue( "\t%i\t%i\t%s", *name, nameType, (char*)value );
        break;

    case 1: // Ansprechen ueber Name
        if ( (strlen( (const char*)name ) < 1) || (strlen( (const char*)name ) > 17) )
        {
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }

        channelBridge->logContinue( "\t%s\t%i\t%s", (char*)name, nameType, (char*)value );
        break;

    case 2: // Ansprechen ueber Systemname
        if ( strlen( (const char*)name ) < 1 )
        {
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        // Bereichspruefung der Systemvariablen
        if ( (strstr( (char*)name, "SERVICE.DBLINPUT" ) && (atoi( (char*)value ) < 0 || atoi( (char*)value ) > 65499)) ||
            (strstr( (char*)name, "SERVICE.DEMANDTIMER" ) && (atoi( (char*)value ) < 0 || atoi( (char*)value ) > 255)) )
        {
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        channelBridge->logContinue( "\t%s\t%i\t%s", (char*)name, nameType, (char*)value );
        break;

    default:
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
    }

    if ( channelBridge->getLastErrorNumber( l_device_address ) != IV_ERROR_SUCCESS )
    {
        if ( nameType == 0 )
        {
            channelBridge->logEnd( "\t%i\t%i\t\t%i\t(%i, %i)", *name, nameType, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        }
        else
        {
            channelBridge->logEnd( "\t%s\t%i\t\t%i\t(%i, %i)", (char*)name, nameType, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        }

        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->setGlobVar( l_device_address, name, nameType, value, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetGlobVar( int connectionNumber, int deviceAddress, const char *name, int nameType, char *value, int valueLength, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || name == nullptr || value == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t\t%i\t\t%i\t%i\t(%i, %i)", nameType, valueLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Bereich pruefen.
    switch ( nameType )
    {
    case 0: // Ansprechen ueber Nummer
        if ( *name < 1 || *name > 8 )
        {
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }

        channelBridge->logContinue( "\t%i\t%i", *name, nameType );
        break;

    case 1: // Ansprechen ueber Name
        if ( (strlen( (const char*)name ) < 1) || (strlen( (const char*)name ) > 17) )
        {
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }

        channelBridge->logContinue( "\t%s\t%i", (char*)name, nameType );
        break;

    case 2: // Ansprechen ueber Systemname
        if ( strlen( (const char*)name ) < 1 )
        {
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }

        channelBridge->logContinue( "\t%s\t%i", (char*)name, nameType );
        break;

    default:
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        break;
    }

    if ( channelBridge->getLastErrorNumber( l_device_address ) != IV_ERROR_SUCCESS )
    {
        if ( nameType == 0 )
        {
            channelBridge->logEnd( "\t%i\t%i\t\t%i\t%i\t(%i, %i)", *name, nameType, valueLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        }
        else
        {
            channelBridge->logEnd( "\t%s\t%i\t\t%i\t%i\t(%i, %i)", (char*)name, nameType, valueLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        }

        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t\t%i\t%i\t(%i, %i)", valueLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->getGlobVar( l_device_address, name, nameType, value, valueLength, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", (char*)value, valueLength, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCCloseRelay( int connectionNumber, int deviceAddress, int number, int timeout, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t%i", number, timeout );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || number < 0 || number > 40 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->closeRelay( l_device_address, number, timeout, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetRelayState( int connectionNumber, int deviceAddress, int number, int *state, int *timeout, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", number );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || number < 0 || number > 40 || state == nullptr || timeout == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->getRelayState( l_device_address, number, state, timeout, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t%i\t%i\t(%i, %i)", *state, *timeout, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCOpenRelay( int connectionNumber, int deviceAddress, int number, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", number );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || number < 0 || number > 40 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->openRelay( l_device_address, number, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetInfo( int connectionNumber, int deviceAddress, const char *name, const char *params, char *text, int *textLength, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || name == nullptr || params == nullptr || text == nullptr || textLength == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%s\t%s\t%s\t%i\t%i\t(%i, %i)", (name == nullptr) ? "" : name, (params == nullptr) ? "" : params,
            (text == nullptr) ? "" : text, (textLength == nullptr) ? 0 : *textLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );

        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%s\t%s\t%s\t%i\t%i\t(%i, %i)", name, params, text, *textLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    std::string info;
    int iRes = channelBridge->getInfo( l_device_address, name, params, info );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        *textLength = static_cast<int>( info.copy( text, *textLength - 1 ) ) + 1;
        text[*textLength - 1] = 0; // Nullterminierung
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 6:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_ENTRANCE2_OFF, errorNumber );
                break;

            case 7:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_MODULE_BUSY, errorNumber );
                break;

            case 8:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_MODULE_ABSENCE, errorNumber );
                break;

            case 28:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
                break;

            case 29:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
                break;

            default:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%s\t%s\t%i\t%i\t(%i, %i)", name, params, text, *textLength, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCOpenComServerMode( int connectionNumber, int deviceAddress, const char *comm, int baudRate, int byteSize, int parity, int stopBits, int flags, int timeoutToClose, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen // HK erweitert um 57600 und 115200
    if ( l_device_address != 254 || comm == nullptr
        || (baudRate != 9600 && baudRate != 19200 && baudRate != 38400 && baudRate != 57600 && baudRate != 115200)
        || flags & 0xFFFF0000 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t%i\t0x%04X\t%i\t%i\t(%i, %i)", 
            (comm == nullptr) ? "" : comm, baudRate, byteSize, parity, stopBits, flags, timeoutToClose, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int iRes = channelBridge->openComServerMode( l_device_address, comm, baudRate, byteSize, parity, stopBits, flags, timeoutToClose );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1: // Es wurden zu wenige Paramteter angegeben.
            case 2: // Es wurden zu viele Paramteter angegeben.
            case 3: // Die Angabe des comm Parameters wird nicht unterstuetzt.
            case 4: // Die Angabe des baudRate Parameters wird nicht unterstuetzt.
            case 5: // Die Angabe des byteSize Parameters wird nicht unterstuetzt.
            case 6: // Die Angabe des parity Paramters wird nicht unterstuetzt.
            case 7: // Die Angabe des stopBits Paramters wird nicht unterstuetzt.
            case 8: // Die Angabe des flags Parameters wird nicht unterstuetzt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
                break;

            case 28: // Passwortangabe fehlerhaft
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
                break;

            case 29: // 3 malige flascheingabe, Geraet geblockt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
                break;

            default:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t%i\t0x%04X\t%i\t%i\t(%i, %i)", comm, baudRate, byteSize, parity, stopBits, flags, timeoutToClose, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCCloseComServerMode( int connectionNumber, int deviceAddress, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address != 254 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Eine Sekunde Pause, fuer Pufferleerung.
    std::this_thread::sleep_for( std::chrono::seconds( 1 ) );

    int iRes = channelBridge->closeComServerMode( l_device_address );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_OVERFLOW, errorNumber );
                break;

            case 28: // Passwort fehlerhaft
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
                break;

            case 29: // 3 malige flascheingabe, Geraet geblockt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
                break;

            default:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

/*
mask:
0x00000001 External Reader options 0=GIS, 1=PHG
0x00000002 IOModul
0x00000004 TBAdaption
0x00000008 ComServer
0x00000010 TransponderWrite
0x00000020 UpdateBiokey
0x00000040 Http-Protokoll
0x00000080 Signalverarbeitung
0x00000100 Setup und Listen beibehalten.

options: zustand 0 ist off, und 1 on
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetOptionFirmware( int connectionNumber, int deviceAddress, int mask, int options, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t%i", mask, options );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || mask < 0 || mask > 0x100 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int iRes = channelBridge->setOptionFirmware( l_device_address, mask, options );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( (iRes & 0xFF) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
            }
            else if ( (iRes & 0xFF) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Verbindungsobjekt bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd("\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetOptionFirmware( int connectionNumber, int deviceAddress, int *mask, int *options, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( "%s\t%i\t%i", __func__, connectionNumber, deviceAddress );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || mask == nullptr || options == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t%i\t%i\t(%i, %i)", (mask == nullptr) ? 0 : mask, (options == nullptr) ? 0 : options, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        
        channelBridge->logEnd( "\t%i\t%i\t%i\t(%i, %i)", mask, options, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int iRes = channelBridge->getOptionFirmware( l_device_address, *mask, *options );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten
            if ( (iRes & 0xFF) == 28 )
            {
                // Passwort wird benoetigt?
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
            }
            else if ( (iRes & 0xFF) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
            }
            else
            {
                // Funktionsfehler : Fehler in Verarbeitung.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t%i\t%i\t(%i, %i)", mask, options, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetFontType( int connectionNumber, int deviceAddress, int type, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", type );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || type < 0 || type > 6 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->setTypeFont( l_device_address, type, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCPressVirtualKey( int connectionNumber, int deviceAddress, int keyCode, int flags, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t%i", keyCode, flags );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || keyCode < 0x70 || (keyCode > 0x7E && keyCode != 0xEE) || flags > 1 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->pressVirtualKey( l_device_address, keyCode, flags, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber  );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetPasswordKey( int connectionNumber, int deviceAddress, char *value, int *valueLength, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || value == nullptr || valueLength == nullptr || *valueLength < 17 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", (value == nullptr) ? "" : value, (valueLength == nullptr) ? 0 : *valueLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", value, *valueLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int iRes = channelBridge->getPasswordKey( l_device_address, value );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        *valueLength = (int)strlen( value );
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        } 
        else
        {
            // Funktionsfehler auswerten.
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", value, *valueLength, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetPassword( int connectionNumber, const char *value, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, "" );

    // zulaessige Bereiche pruefen
    if ( value == nullptr || strlen( value ) > 16 )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (value == nullptr) ? "" : "****", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS, errorNumber );

    channelBridge->setPassword( value );

    channelBridge->logEnd( "\t****\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 1, channelBridge->getLastErrorNumber( 0 ) );

    return 1;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCReset( int connectionNumber, int deviceAddress, int mode, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", mode );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    if ( l_device_address == 0 
        || (mode != 1 && mode != 2 && mode != 3 && mode != 6 && mode != 7 && mode != 100 && mode != 101) )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(0, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->reset( l_device_address, mode, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetFlashStatus( int connectionNumber, int deviceAddress, int *value, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || value == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t%i\t(%i, %i)", *value, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->getFlashStatus( l_device_address, value, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t%i\t(%i, %i)", *value, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetCommunicationPassword( int connectionNumber, const char *value, int valueLength, int sessionTimeout, int *errorNumber )
{
    // Auf Logchannel ausweichen, wenn die Passwortangabe fuer alle gueltig sein soll.
    int id = connectionNumber;
    if ( connectionNumber == -1 )
    {
        id = LOG_CHANNEL_ID;
    }

    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBeginPure( __func__, "\t%i", connectionNumber );

    // zulaessige Bereiche pruefen
    if ( value == nullptr || valueLength < 0 || valueLength > 16
        || sessionTimeout < 0 || (sessionTimeout > 0 && sessionTimeout < 60) || sessionTimeout > 100000 )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t(%i, %i)", (value == nullptr) ? "" : "****", valueLength, sessionTimeout, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS, errorNumber );

    int res = theApp.SetCommunicationPassword( connectionNumber, value, valueLength, sessionTimeout );
    if ( res == 0 )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_CHANNEL, errorNumber );
    }

    channelBridge->logEnd( "\t****\t%i\t%i\t%i\t(%i, %i)", valueLength, sessionTimeout, (errorNumber == nullptr) ? 0 : *errorNumber, res, channelBridge->getLastErrorNumber( 0 ) );

    return res;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCReadHardwareInfo( int connectionNumber, int deviceAddress, char *value, int *valueLength, int flags, int *errorNumber ) 
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || value == nullptr || valueLength == nullptr || flags != 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t(%i, %i)", (value == nullptr) ? "" : value, (valueLength == nullptr) ? 0 : *valueLength, 
            flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );

        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t(%i, %i)", value, *valueLength, flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int res = 0;
    unsigned char ucVal;
    unsigned int uiVal;
    unsigned short usVal;
    enum
    {
        ucType, 
        uiType, 
        usType
    } putType;

    // Klasse fuer die Verwaltung der Daten der Hardwareinfo
    ini_struct is;
    // Leeres PropertyTree Objekt
    using boost::property_tree::ptree;
    ptree pt;
    int readAll = 0;

    // Laden der vorliegenden Ini-Datei.
    // Ausnahme wird ausgeloest, wenn ein Fehler vorliegt.
    try
    {
        //read_ini(fileName, pt);
        // Aktuell werden noch alle gelesen
        readAll = 1;
    }
    catch ( boost::property_tree::ini_parser::ini_parser_error &e )
    {
        // Oeffnen der Datei fehlgeschlagen?
        if ( e.message() != "cannot open file" )
        {
            //TRACE("Fehlernachricht: %s\n", e.message().c_str());
        }

        // Es werden alle moeglichen Informationen ermittelt.
        readAll = 1;
    }

    HardwareInfo hardwareInfo;
    if ( (res = channelBridge->getHardwareInfo( l_device_address, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_UNREACHABLE, errorNumber );
        res = 0;
    }
    else
    {
        // Aktuell auf die "Vorgängerversion" mappen.
        StructHardwareInfo hardwareInfoOld( hardwareInfo.exportStructHardwareInfo() );

        std::set<std::string>::iterator i;
        i = is.setting.begin();
        for ( ; i != is.setting.end(); ++i )
        {

            if ( readAll == 0 && pt.get_child_optional( *i ) == boost::none ) // Dieser Key soll nicht gelesen werden.
                continue;

            res = IV_ERROR_SUCCESS;
            putType = ucType;

            if ( *i == "serial" )
            {
                if ( (res = channelBridge->m_channel4.readSerialNumber( l_device_address, uiVal )) < IV_ERROR_SUCCESS ) 
                    break;

                putType = uiType;
            }
            else if ( *i == "version" )
            {
                // Version der Hardwareinfo ermitteln
                ucVal = hardwareInfoOld.version;
            }
            else if ( *i == "type" )
            {
                // Geraetetyp ermitteln
                ucVal = hardwareInfoOld.typeDevice;
            }
            else if ( *i == "board" )
            {
                // Boardtyp ermitteln
                ucVal = hardwareInfoOld.typeBoard;
            }
            else if ( *i == "rs485" )
            {
                // RS485 ermitteln
                ucVal = hardwareInfoOld.componentRS485_0;
            }
            else if ( *i == "tcpip" )
            {
                // TCPIP ermitteln
                ucVal = hardwareInfoOld.typeTCPIP;
            }
            else if ( *i == "modemMC" )
            {
                // ModemMC35i ermitteln
                ucVal = hardwareInfoOld.typeModemMC;
            }
            else if ( *i == "rs485ZK" )
            {
                // RS485 (intern) ermitteln
                ucVal = hardwareInfoOld.componentRS485_1;
            }
            else if ( *i == "transponder" )
            {
                // Typ ermitteln
                ucVal = hardwareInfoOld.typeTransponder;
            }
            else if ( *i == "display" )
            {
                // Typ ermitteln (erst ab 04.01.01.p / Bootloader 4.0)
                ucVal = hardwareInfoOld.typeDisplay;
            }
            else if ( *i == "keypad" )
            {
                // Typ ermitteln
                ucVal = hardwareInfoOld.typeKeypad;
            }
            else if ( *i == "mmc" )
            {
                // Typ ermitteln
                ucVal = hardwareInfoOld.typeMMC;
            }
            else if ( *i == "gps" )
            {
                // Typ ermitteln
                ucVal = hardwareInfoOld.typeGPS;
            }
            else if( *i == "usb" )
            {
                // Komponende ermitteln
                ucVal = hardwareInfoOld.componentUSB;
            }
            else if ( *i == "radio" )
            {
                // Typ ermitteln
                ucVal = hardwareInfoOld.typeRadio;
            }
            else if ( *i == "dcf77" )
            {
                // Typ ermitteln
                ucVal = hardwareInfoOld.typeDCF77;
            }
            else if ( *i == "inAnalog" )
            {
                // Komponende ermitteln
                uiVal = hardwareInfoOld.componentInAnalog;

                putType = uiType;
            }
            else if ( *i == "inDigital" )
            {
                // Komponende ermitteln
                uiVal = hardwareInfoOld.componentInDigital;

                putType = uiType;
            }
            else if ( *i == "outDigital" )
            {
                // Komponende ermitteln
                uiVal = hardwareInfoOld.componentOutDigital;

                putType = uiType;
            }
            else if ( *i == "fingerprint" )
            {
                // Typ ermitteln
                ucVal = hardwareInfoOld.typeFingerprint;
            }
            else if ( *i == "externalReader" )
            {
                // Typ ermitteln
                if ((res = channelBridge->m_channel4.readExternalTransponderReaderType(l_device_address, ucVal)) < IV_ERROR_SUCCESS)
                    break;
            }
            else if ( *i == "ioModule" )
            {
                // Option ermitteln
                if ((res = channelBridge->m_channel4.readIoModuleOption(l_device_address, ucVal)) < IV_ERROR_SUCCESS)
                    break;
            }
            else if ( *i == "tbAdaption" )
            {
                // Option ermitteln
                if ((res = channelBridge->m_channel4.readTimeboyAdaptionOption(l_device_address, ucVal)) < IV_ERROR_SUCCESS)
                    break;
            }
            else if ( *i == "comServer" )
            {
                // Option ermitteln
                if ((res = channelBridge->m_channel4.readComServerOption(l_device_address, ucVal)) < IV_ERROR_SUCCESS)
                    break;
            }
            else if ( *i == "tpWrite" )
            {
                // Option ermitteln
                if ((res = channelBridge->m_channel4.readTransponderWriteOption(l_device_address, ucVal)) < IV_ERROR_SUCCESS)
                    break;
            }
            else if ( *i == "http" )
            {
                // Option ermitteln
                if ( (res = channelBridge->m_channel4.readHttpOption( l_device_address, ucVal )) < IV_ERROR_SUCCESS)
                    break;
            }
            else if ( *i == "language" )
            {
                // Option ermitteln
                if ( (res = channelBridge->m_channel4.readLocalizationOption( l_device_address, ucVal )) < IV_ERROR_SUCCESS)
                    break;
            }
            else if ( *i == "fingerprintUpdate" )
            {
                // Option ermitteln
                if ( (res = channelBridge->m_channel4.readFingerprintScannerUpdateOption( l_device_address, ucVal )) < IV_ERROR_SUCCESS)
                    break;
            }
            else if ( *i == "setup" )
            {
                unsigned short mask = 0;
                // Option ermitteln
                if ( (res = channelBridge->m_channel4.readSetupOptions( l_device_address, mask, usVal )) < IV_ERROR_SUCCESS)
                    break;

                putType = usType;
            }
            else if ( *i == "touch" )
            {
                // Typ ermitteln
                ucVal = hardwareInfoOld.typeTouch;
            }
            else if ( *i == "sensor" )
            {
                // Typ ermitteln
                ucVal = hardwareInfoOld.typeSensor;
            }
            else if ( *i == "battery" )
            {
                // Komponente ermitteln
                ucVal = hardwareInfoOld.componentBattery;
            }
            else if( *i == "charge" )
            {
                // Komponente ermitteln
                ucVal = hardwareInfoOld.componentCharge;
            }
            else
            {
                // Kein bekannter Eintrag.
                continue;
            }

            if ( res == IV_ERROR_SUCCESS ) 
            {
                switch ( putType )
                {
                case uiType:
                    pt.put( *i, uiVal );
                    break;

                case usType:
                    pt.put( *i, usVal );
                    break;

                default:
                    pt.put( *i, ucVal );
                }

                //TRACE("Key: %s\n", (*i).c_str());
            }
        }

        try
        {
            // Write property tree to INI file
            std::ostringstream os;
            write_ini( os, pt );
            os << std::ends;
            strncpy( value, os.str().data(), *valueLength - 1 );
            *valueLength = static_cast<int>( strlen( value ) );
            res = 1;
        }
        catch( std::exception &e )
        {
            (void)e;
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FILE_WRITE, errorNumber );
            res = 0;
        }
    }


    channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t(%i, %i)", value, *valueLength, flags, (errorNumber == nullptr) ? 0 : *errorNumber, res, channelBridge->getLastErrorNumber( l_device_address ) );
    return res;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCRecordVolume( int connectionNumber, int deviceAddress, int *recordCount, int *memoryUsage, int *memorySize, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || recordCount == nullptr || memoryUsage == nullptr || memorySize == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t%i\t%i\t%i\t(%i, %i)", (recordCount == nullptr) ? 0 : *recordCount, (memoryUsage == nullptr) ? 0 : *memoryUsage, (memorySize == nullptr) ? 0 : *memorySize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t%i\t%i\t%i\t(%i, %i)", *recordCount, *memoryUsage, *memorySize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int iRes = channelBridge->recordVolume( l_device_address, *recordCount, *memoryUsage, *memorySize );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        } 
        else
        {
            // Funktionsfehler auswerten.
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t%i\t%i\t%i\t(%i, %i)", *recordCount, *memoryUsage, *memorySize, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCCheckAE( int connectionNumber, int deviceAddress )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->checkDevice( l_device_address, &l_error_number, 1 );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE void* DF_API_CALL DFCGetClass( int connectionNumber, int classId )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return nullptr;
    }

    switch ( classId )
    {
    case 0:
        return channelBridge;

    case 1:
    case 2:
        return &channelBridge->m_channel1_2;

    case 4:
        return &channelBridge->m_channel4;

    default:
        return nullptr;
    }

    return nullptr;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCCommandInspectionStation( int connectionNumber, int deviceAddress, int requestAddress, char *requestParams, int requestParamsSize, int *answerAddress, char *answerParams, int *answerParamsSize, int secondsToWaitForAnswer )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return -1;
    }

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // Parameter pruefen
    if (secondsToWaitForAnswer < 1 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );

        return -IV_ERROR_FUNCTION_PARAMETER;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    return channelBridge->m_channel4.inspectionStationCommand( l_device_address, requestAddress, requestParams, requestParamsSize, (unsigned char&)*answerAddress, answerParams, (int&)*answerParamsSize, secondsToWaitForAnswer );
}

#ifdef _DEBUG

DF_API_EXTERN_DECORATE int DF_API_CALL DFCWatchdogReset( int connectionNumber, int deviceAddress )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );
    int l_error_number = IV_ERROR_SUCCESS;

    int iRes = channelBridge->restart( l_device_address, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

#endif //_DEBUG

DF_API_EXTERN_DECORATE int DF_API_CALL DFCCustomCommand( int connectionNumber, int deviceAddress, int identification, int customCommand, const char *dataIn, int sizeIn, char *dataOut, int *sizeOut, int *errorNumber )
{
	ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
	// Wurde Channel-Objekt ermittelt.
	if ( channelBridge == nullptr )
	{
		if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

		return 0;
	}

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", customCommand );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

	// zulaessige Bereiche pruefen
	if ( l_device_address == 0
		|| customCommand < 0
		|| dataIn == nullptr || dataOut == nullptr || sizeOut == nullptr )
	{
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%s\t%i\t%s\t%i\t%i\t(%i, %i)", (dataIn == nullptr) ? "" : dataIn, sizeIn, (dataOut == nullptr) ? "" : dataOut, (sizeOut == nullptr) ? 0 : *sizeOut, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
		return 0;
	}

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%s\t%i\t%s\t%i\t%i\t(%i, %i)", dataIn, sizeIn, dataOut, *sizeOut, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
		return 0;
	}

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

	int iRes = channelBridge->customCommand( l_device_address, identification, customCommand, dataIn, sizeIn, dataOut, sizeOut );
	if ( iRes == IV_ERROR_SUCCESS )
	{
		// Kein Fehler
		iRes = 1;
	}
	else if ( iRes > IV_ERROR_SUCCESS )
	{
		// Fehler bei Durchfuehrung des Befehls
		// Modulfehler auswerten.
		if ( (iRes >> 8) & 0xFF )
		{
			if ( ((iRes >> 8) & 0xFF) == 2 )
			{
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
			}
			else
			{
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
			}
		}
		else
		{
			// Funktionsfehler auswerten.
			switch ( iRes & 0xFF )
			{
			case 28:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
				break;

			case 29:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
				break;

			default:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
			}
		}

		iRes = 0;
	}
	else
	{
		// Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
		iRes = 0;
	}

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t%s\t%i\t%i\t(%i, %i)", dataIn, sizeIn, dataOut, *sizeOut, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

	return iRes;
}

DF_API_EXTERN int DF_API_CALL DFCGetLastErrorNumber( int connectionNumber, int deviceAddress )
{
    int res = IV_ERROR_SUCCESS;
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return IV_ERROR_CHANNEL;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    res = channelBridge->getLastErrorNumber( l_device_address );

    channelBridge->logEnd("\t(%i, %i)", res, channelBridge->getLastErrorNumber( l_device_address ) );

	return res;
}

DF_API_EXTERN int DF_API_CALL DFCBlockTransferSetDuration( int connectionNumber, int deviceAddress, int blockTypesAllowed, int duration )
{
    int res = 0;
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return res;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", duration );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );
    if ( deviceAddress == -1 )
    {
        l_device_address = 255;
    }

    // zulaessige Bereiche pruefen
    if ( (l_device_address != 255 && l_device_address == 0)
        || (blockTypesAllowed & ControllerSegment::btt_mask_known) == 0
        || (blockTypesAllowed & ~ControllerSegment::btt_mask_known) != 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", res, channelBridge->getLastErrorNumber( l_device_address ) );
        return res;
    }

    //TODO: Wert darf nicht auf 0 gesetzt werden, wenn aktuell Aktiv!!
    if ( (l_device_address == 255 || blockTypesAllowed == ControllerSegment::btt_none || duration == 0)
        && channelBridge->getBlockTransferState(l_device_address) == ControllerSegment::BlockTransferState::active )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_BLOCK_TRANSFER_ACTIVE );
        res = 0;
    }
    else
    {
        channelBridge->setBlockTransferSettings( l_device_address, blockTypesAllowed, duration );
        res = 1;
    }

    channelBridge->logEnd("\t(%i, %i)", res, channelBridge->getLastErrorNumber(l_device_address));

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCBlockTransferResume( int connectionNumber, int deviceAddress )
{
    int res = 0;
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return res;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", res, channelBridge->getLastErrorNumber( l_device_address ) );
        return res;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", res, channelBridge->getLastErrorNumber( l_device_address ) );
        return res;
    }

    res = channelBridge->blockTransferResume( l_device_address );
    if ( res < IV_ERROR_SUCCESS )
    {
        channelBridge->setLastErrorNumber( l_device_address, -res );
        res = 0;

        if ( channelBridge->getLastErrorNumber( l_device_address ) != IV_ERROR_BLOCK_TRANSFER_BREAK )
        {
            channelBridge->blockTransferDiscard( l_device_address );
        }
    }
    else if ( res > IV_ERROR_SUCCESS )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING );
    }
    else
    {
        res = 1;
    }

    channelBridge->logEnd( "\t(%i, %i)", res, channelBridge->getLastErrorNumber( l_device_address ) );

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCBlockTransferGetState( int connectionNumber, int deviceAddress, int *currentType, int *currentState )
{
    int res = 0;
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return res;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || currentType == nullptr || currentState == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t\t\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return res;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t\t\t(%i, %i)", res, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    if ( channelBridge->getBlockTransferState( l_device_address ) == ControllerSegment::BlockTransferState::inactive )
    {
        *currentType = ControllerSegment::btt_none;
        *currentState = static_cast<int32_t>( ControllerSegment::BlockTransferState::inactive );
    }
    else
    {
        *currentType = channelBridge->getBlockTransferType( l_device_address );
        *currentState = static_cast<int32_t>( channelBridge->getBlockTransferState( l_device_address ) );
    }

    res = 1;
    channelBridge->logEnd( "\t%i\t%i\t(%i, %i)", *currentType, *currentState, res, channelBridge->getLastErrorNumber( l_device_address ) );

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCBlockTransferDiscard( int connectionNumber, int deviceAddress )
{
    int res = 0;
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return res;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", res, channelBridge->getLastErrorNumber( l_device_address ) );
        return res;
    }

    res = channelBridge->blockTransferDiscard( l_device_address );
    if ( res < IV_ERROR_SUCCESS )
    {
        channelBridge->setLastErrorNumber( l_device_address, -res );
        res = 0;
    }
    else if ( res > IV_ERROR_SUCCESS )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING );
    }
    else
    {
        res = 1;
    }

    channelBridge->logEnd( "\t(%i, %i)", res, channelBridge->getLastErrorNumber( l_device_address ) );

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCReadDataBlock( int connectionNumber, int deviceAddress, const char *name, char *data, int *length, int flags )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || name == nullptr || data == nullptr || length == nullptr || flags > 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%s\t%p\t%i\t%i\t(%i, %i)", (name == nullptr) ? "" : name, (data == nullptr) ? "" : data, (length == nullptr) ? 0 : *length, flags, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t%p\t%i\t%i\t(%i, %i)", name, data, *length, flags, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    // Datenblock lesen
    uint32_t l_length = (static_cast<unsigned int>(*length) > 0xFFFFFFFF) ? 0xFFFFFFFF : static_cast<uint32_t>(*length);
    int iRes = channelBridge->read_data_block( l_device_address, name, data, &l_length, flags );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        *length = static_cast<int>( l_length );

        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING );
            }
        }
        else
        {
            // Modulfehler auswerten.
            channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING );
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }

    channelBridge->logEnd( "\t%s\t%p\t%i\t%i\t(%i, %i)", name, data, *length, flags, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}
