/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/
#pragma once
#ifndef DATAFOX_COMMUNICATION_PROTOCOL_CRYPT_H
#define DATAFOX_COMMUNICATION_PROTOCOL_CRYPT_H

// Basisklasse einbinden.
#include "datafox/communication/protocol_base.h"
#include <time.h>
#include <string>

#include "common/crc.h"
#include "datafox/communication/aes_crypt_session.h"

#define IV_MODULE_PROTOCOL_CRYPT                 0x0000
#define IV_FUNCTION_PROTOCOL_PING                0x0000
#define IV_FUNCTION_EXCHANGE_PUBLIC_KEY          0x0046

namespace Datafox
{
    namespace Communication
    {
        /// Basisklasse fuer pure protokollbasierte Kommunikationsaufgaben.
        class ProtocolCrypt : public ProtocolBase
        {
        public:
            ProtocolCrypt( int32_t connectionNumber = ConnectionBase::defaultConnectionNumber(), uint8_t deviceAddress = defaultDeviceAddress() );
            virtual ~ProtocolCrypt();

            // Speicher fuer Pakete/Daten
        public:
            void reset();

        public:
            uint8_t m_packetNumberPrevious;
            uint8_t m_packetDataPrevious[IV_PACKET_LENGTH_DATA];
            uint16_t m_packetDataSizePrevious;

        protected:
            AesCryptSession m_aesCryptSession[32];
        public:
            inline uint8_t currentSessionNumber()
            {
                return ((deviceNumberReceiver() < 1 || deviceNumberReceiver() > 31) ? 0 : deviceNumberReceiver());
            }
            inline AesCryptSession &aesCryptSession()
            {
                return m_aesCryptSession[currentSessionNumber()];
            }

        public:
            // Passwort fuer die verschluesselte Kommunikation setzen
            int32_t setCommunicationPassword( const char* password, uint8_t size, uint32_t sessionTimeout, bool global );

        public:
            // Ausfuehren eines Protokollbefehls
            virtual int32_t call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int32_t contextValue = -1 );

        protected:
            // Paketdaten ggf. verschluesseln.
            virtual int32_t writeProtocolCrypt();
        public:
            // Streamdaten ggf. verschluesseln.
            int writeStreamCrypt( std::string &buffer );
        protected:
            // Packetdaten ggf. entschluesseln.
            virtual int32_t readProtocolCrypt();
        public:
            // Streamdaten ggf. entschluesseln.
            virtual int32_t readStreamCrypt( std::string &buffer );

            // Pruefen ob aktuell eine Verschluesselte Session aktiv ist.
            bool isCrypt()
            {
                return (aesCryptSession().m_cryptMode > 1);
            }

        public:
            // Austausch eines Session-Key fuer die verschluesselte Kommunikation
            int32_t exchangeSessionKey( uint8_t device_address );

        };

    } // namespace Communication
} // namespace Datafox

#endif
